/* vaixells.c
 * Autors: Sergio Blanco Cuaresma
 *         Ramn Quionero Bonmat
 *         Valentn Rey Gmez
 * Data: Gener del 2002
 * Descripci: Execuci: ./vaixells nom_jugador 1 1 fitxer.txt
 * A on el primer 1 significa que comena a jugar el programa (0 comena el jugador)
 * el segon 1 significa que mostri missatges per la sortida d'error (0 que no)
 * i fitxer.txt es el nom d'un fitxer que cont les posicions dels vaixells del programa.
 *
 * El programa juga al joc dels vaixells contra l'usuari en una taula de 10x10, fent tirs
 * del estil A1, B5... espera rebre A, T o E segons si ha sigut Aigua, Tocat o enfonsat, i
 * tb llegeix els tirs de l'usuari retornant tb A, T o E. Si algun dels 2 ha tocat o enfonsat
 * un vaixell, continua disparant
 *
 * Els tirs realitzats pel programa son generats de forma aleatoria pero inteligenment, si
 * fa blanc, els proxims tirs els realitza al voltant. I si enfonsa un vaixell, no fa cap tir
 * al voltant d'aquest ja que es supossa que no poden haber-hi 2 vaixells continus.
 *
 */

#include <stdio.h>
#include <fcntl.h>
#include <assert.h>
#include <errno.h>
#include "tipus.h"
#include <stdlib.h>
#include <time.h>

#define N 10

#define BUIT 'B'
#define VAIXELL 'V'

#define AIGUA 'A'
#define TOCAT 'T'
#define ENFONSAT 'E'

#define POSLL(n) (n-'A')
#define POSN(n) (n-'0')
#define LLPOS(n) (n+'A')
#define NPOS(n) (n+'0')

struct vaixell_transatlantic transatlantic;
struct vaixell_buque buque[2];
struct vaixell_yot yot[3];
struct vaixell_barca barca[4];
BOOLEAN tirs[10][10], tirs_aux[10][10];

// Llegeix un transatlantic
void llegir_transatlantic (FILE * posicions)
{
  char bucle;

  for (bucle = 0; bucle < 4; bucle++) {
    transatlantic.pos[bucle].lletra = toupper (fgetc (posicions));
    transatlantic.pos[bucle].numero = POSN (fgetc (posicions));
    transatlantic.pos[bucle].tocat = FALSE;
  }
  transatlantic.enfonsat = FALSE;       // Inicialitzem
  fgetc (posicions);            // End of line
}

// LLegueix dos buques
void llegir_buques (FILE * posicions)
{
  char bucle1, bucle2;

  for (bucle1 = 0; bucle1 < 2; bucle1++)        // Per cada buque
  {
    for (bucle2 = 0; bucle2 < 3; bucle2++)      // Per cada posici
    {
      buque[bucle1].pos[bucle2].lletra = toupper(fgetc (posicions));
      buque[bucle1].pos[bucle2].numero = POSN (fgetc (posicions));
      buque[bucle1].pos[bucle2].tocat = FALSE;
    }
    buque[bucle1].enfonsat = FALSE;     // Inicialitzem
    fgetc (posicions);          // End of line
  }
}

// Llegueix tres yots
void llegir_yots (FILE * posicions)
{
  char bucle1, bucle2;

  for (bucle1 = 0; bucle1 < 3; bucle1++)        // Per cada yot
  {
    for (bucle2 = 0; bucle2 < 2; bucle2++)      // Per cada posici
    {
      yot[bucle1].pos[bucle2].lletra = toupper(fgetc (posicions));
      yot[bucle1].pos[bucle2].numero = POSN (fgetc (posicions));
      yot[bucle1].pos[bucle2].tocat = FALSE;
    }
    yot[bucle1].enfonsat = FALSE;       // Inicialitzem
    fgetc (posicions);          // End of line
  }
}

// Llegueix quatre barques
void llegir_barques (FILE * posicions)
{
  char bucle1, bucle2;

  for (bucle1 = 0; bucle1 < 4; bucle1++)        // Per cada barca
  {
    barca[bucle1].pos.lletra = toupper(fgetc (posicions));
    barca[bucle1].pos.numero = POSN (fgetc (posicions));
    barca[bucle1].pos.tocat = FALSE;
    barca[bucle1].enfonsat = FALSE;
    getc (posicions);           // End of line
  }
}


// Llegueix totes les posicions del vaixells
void llegir_fitxer (char *fitxer)
{
  FILE *posicions;

  if ((posicions = fopen (fitxer, "r")) == NULL) {
    perror ("posicions.txt");
    exit (1);
  }

  llegir_transatlantic (posicions);
  llegir_buques (posicions);
  llegir_yots (posicions);
  llegir_barques (posicions);

  fclose (posicions);
}

// Comprova si ens han tocat, enfonsat o fet aigua a partir d'una posici indicada com argument
char comprovar_tir (char *resposta_tir)
{
  char bucle, bucle2;

  // Si ja esta enfonsat no cal comprovar si ha sigut tocat
  if (transatlantic.enfonsat == FALSE) {
    bucle = 0;
    // Mentre no trobem cap posici igual, avancem
    while (((transatlantic.pos[bucle].numero != POSN (resposta_tir[1]))
            || (transatlantic.pos[bucle].lletra != resposta_tir[0]))
           && (bucle < 3)) {
      bucle++;
    }
    // Si hem sortit pq hi ha una posici igual
    if ((transatlantic.pos[bucle].numero == POSN (resposta_tir[1]))
        && (transatlantic.pos[bucle].lletra == resposta_tir[0])) {
      transatlantic.pos[bucle].tocat = TRUE;    // Ens han tocat
      // Si totes les posicions estan tocades
      if ((transatlantic.pos[0].tocat == TRUE) && (transatlantic.pos[1].tocat == TRUE) && (transatlantic.pos[2].tocat == TRUE) && (transatlantic.pos[3].tocat == TRUE)) {
        transatlantic.enfonsat = TRUE;  // Ens han enfonsat
        return ENFONSAT;
      }
      return TOCAT;
    }
  }
  // Per cada buque (2 en total)
  for (bucle2 = 0; bucle2 < 2; bucle2++) {
    // Si ja esta enfonsat no cal comprovar si ha sigut tocat
    if (buque[bucle2].enfonsat == FALSE) {
      bucle = 0;
      // Mentre no trobem cap posici igual, avancem
      while (((buque[bucle2].pos[bucle].numero != POSN (resposta_tir[1]))
              || (buque[bucle2].pos[bucle].lletra != resposta_tir[0]))
             && (bucle < 2)) {
        bucle++;
      }
      // Si hem sortit pq hi ha una posici igual
      if ((buque[bucle2].pos[bucle].numero == POSN (resposta_tir[1]))
          && (buque[bucle2].pos[bucle].lletra == resposta_tir[0])) {
        buque[bucle2].pos[bucle].tocat = TRUE;  // Ens han toca
        // Si totes les posicions estan tocades
        if ((buque[bucle2].pos[0].tocat == TRUE) && (buque[bucle2].pos[1].tocat == TRUE) && (buque[bucle2].pos[2].tocat == TRUE)) {
          buque[bucle2].enfonsat = TRUE;        // Ens han enfonsat
          return ENFONSAT;
        }
        return TOCAT;
      }
    }
  }

  // Per cada yot (3 en total)
  for (bucle2 = 0; bucle2 < 3; bucle2++) {
    // Si ja esta enfonsat no cal comprovar si ha sigut tocat
    if (yot[bucle2].enfonsat == FALSE) {

      // Comprovem si es tracta de la primera posici
      if ((yot[bucle2].pos[0].numero == POSN (resposta_tir[1]))
          && (yot[bucle2].pos[0].lletra == resposta_tir[0])) {
        yot[bucle2].pos[0].tocat = TRUE;        // Ens han toca
        if ((yot[bucle2].pos[0].tocat == TRUE) && (yot[bucle2].pos[1].tocat == TRUE)) {
          yot[bucle2].enfonsat = TRUE;  // Ens han enfonsat
          return ENFONSAT;
        }
        return TOCAT;
      }
      // Comprovem si es tracta de la segona posici
      if ((yot[bucle2].pos[1].numero == POSN (resposta_tir[1]))
          && (yot[bucle2].pos[1].lletra == resposta_tir[0])) {
        yot[bucle2].pos[1].tocat = TRUE;        // Ens han toca
        if ((yot[bucle2].pos[0].tocat == TRUE) && (yot[bucle2].pos[1].tocat == TRUE)) {
          yot[bucle2].enfonsat = TRUE;  // Ens han enfonsat
          return ENFONSAT;
        }
        return TOCAT;
      }

    }
  }

  // Per cada barca (4 en total)
  for (bucle2 = 0; bucle2 < 4; bucle2++) {
    // Si ja esta enfonsat no cal comprovar si ha sigut tocat// Si ja esta enfonsat no cal comprovar si ha sigut tocat
    if (barca[bucle2].enfonsat == FALSE) {
      if ((barca[bucle2].pos.numero == POSN (resposta_tir[1]))
          && (barca[bucle2].pos.lletra == resposta_tir[0])) {
        barca[bucle2].pos.tocat = TRUE; // Tocat
        barca[bucle2].enfonsat = TRUE;  // i enfonsat
        return ENFONSAT;
      }
    }
  }
  // Si no ha retornat cap enfonsat o tocat en aquest punt
  // llavors es tracta d'un lloc a on no hi ha cap vaixell o
  // ja ha sigut tocat abans (situaci que no s'hauria de donar si
  // el programa juga amb si mateix degut al disseny)
  return AIGUA;
}


// Retorna true si tots els vaixells que tenim estan enfonsats, false en cas contrari
BOOLEAN tots_enfonsats (void)
{
  if ((transatlantic.enfonsat) && (buque[0].enfonsat) && (buque[1].enfonsat) && (yot[0].enfonsat) && (yot[1].enfonsat)
      && (yot[2].enfonsat) && (barca[0].enfonsat) && (barca[1].enfonsat) && (barca[2].enfonsat) && (barca[3].enfonsat)) {
    return TRUE;
  } else {
    return FALSE;
  }
}

// Comprova si s'ha realitzat un tir
BOOLEAN tir_ja_fet (char lletra, char numero)
{
  return tirs[POSLL (lletra)][numero];
}

// Guardem la posici a on fem un tir, per no repetir-ho
void afegir_tir (char lletra, char numero)
{
  tirs_aux[POSLL (lletra)][numero] = TRUE;      // Ho guardem tb a la auxiliar per no perdre dades
  // quan fem tirs = tirs_aux
  tirs[POSLL (lletra)][numero] = TRUE;
}

// En tirs_aux marquem la posicio que ens indica i tot el seu voltant, ja
// que al costat del vaixell no pot haver-hi cap altre. (Fins que el vaixell
// no ha sigut enfonsat, el tirs_aux no pasa a ser el tirs principal)
void afegir_tir_aux (char lletra, char numero)
{
  char fila, columna;

  for (fila = -1; fila < 2; fila++) {
    for (columna = -1; columna < 2; columna++) {
      if ((POSLL (lletra) + columna < 10) && (POSLL (lletra) + columna > -1) && (numero + fila < 10) && (numero + fila > -1)) {
        tirs_aux[POSLL (lletra) + columna][numero + fila] = TRUE;
      }
    }
  }
}


// Ens guarda a lletra i numero un tir nou, que no hagui sigut realitzat ja 
void nou_tir (char *lletra, char *numero)
{
  // Posici aleatoria
  *lletra = LLPOS (abs ((char) (rand ()) % 10));
  *numero = abs ((char) (rand ()) % 10);

  while (tir_ja_fet (*lletra, *numero)) {       // Busquem si ja hem fet aquest tir, si es aixi calculem un nou
    *lletra = LLPOS (abs ((char) (rand ()) % 10));
    *numero = abs ((char) (rand ()) % 10);
  }
}

// Cos principal
int main (int argc, char *argv[])
{
  char nom_jugador[20], lletra, numero, resposta, resposta_tir[2], c;
  char begin, err, encerts = 0, offset_lletra = 0, offset_numero = 0, direccio = 0;
  char i, j;
  BOOLEAN proxim = FALSE, lloc_trobat = FALSE;

  setbuf (stdin, NULL);         // anulem el buffer de l'entrada estandard
  setbuf (stdout, NULL);        // anulem el buffer de sortida estandard

  // Tractem arguments
  if (argc != 5) {
    fprintf (stderr, "Error als parmetres\n");
    fprintf (stderr, "\tus: %s nom_jugador torn_maquina sortida_error fitxer_posicions\n", argv[0]);
    exit (1);
  }
  strncpy (nom_jugador, argv[1], 20);
  begin = atoi (argv[2]);
  err = atoi (argv[3]);
  if ((begin != 0) && (begin != 1)) {
    fprintf (stderr, "\n\tError, en els booleans qui_comena i sortida_error\n");
  }
  // Obtenim les posicions dels vaixells de la mquina
  llegir_fitxer (argv[4]);

  // semilla
  //srand(time(NULL));    // Per compatibilitat amb MSDOS
  srand (getpid ());            // Per la part 2-2 de la prctica

  // Mentre no hi hagui cap guanyador
  do {
    if (begin == 1)             // Torn de la mquina
    {
      if (proxim) {             // Hi ha un vaixell proxim a la posici a on ens trobem (ho sabem per un tir que ha fet blanc)
        lloc_trobat = FALSE;
        while (!lloc_trobat) {  // Mentre no calculem un lloc correcte a on disparar
          switch (direccio) {   // Direccio cap a on provarem per veure si esta el vaixell
            case 0:            // Cap a la dreta
              if (offset_lletra + 1 + lletra < 75) {    // Si no ens sortim de la taula (74 = 'J')
                offset_lletra++;
              }
              if (tir_ja_fet (lletra + offset_lletra, numero)) {        // Si ja hem fet aquest tir
                direccio = (direccio + 1) % 4;  // Seguent direccio sense sortir del rang 0-3
                offset_lletra = 0;
              } else {
                lloc_trobat = TRUE;     // Hem trobat lloc a on disparar
              }
              break;
            case 1:            // Cap a l'esquerra
              if (offset_lletra - 1 + lletra > 64) {    // Si no es sortim de la taula (65 = 'A')
                offset_lletra--;
              }
              if (tir_ja_fet (lletra + offset_lletra, numero)) {
                direccio = (direccio + 1) % 4;  // Seguent direccio sense sortir del rang 0-3
                offset_lletra = 0;
              } else {
                lloc_trobat = TRUE;
              }
              break;
            case 2:            // Cap adalt
              if (offset_numero - 1 + numero > -1) {
                offset_numero--;
              }
              if (tir_ja_fet (lletra, numero + offset_numero)) {
                direccio = (direccio + 1) % 4;  // Seguent direccio sense sortir del rang 0-3
                offset_numero = 0;
              } else {
                lloc_trobat = TRUE;
              }
              break;
            case 3:            // Cap abaix
              if (offset_numero + 1 + numero < 10) {
                offset_numero++;
              }
              if (tir_ja_fet (lletra, numero + offset_numero)) {
                direccio = (direccio + 1) % 4;  // Seguent direccio sense sortir del rang 0-3
                offset_numero = 0;
              } else {
                lloc_trobat = TRUE;
              }
              break;
          }
        }                       // Al sortir del while ja tenim una posici a on disparar
        afegir_tir (lletra + offset_lletra, numero + offset_numero);    // Guardem per no tornar a repetir

      } else {                  // Si no hi ha cap vaixell proxim
        offset_numero = offset_lletra = 0;
        nou_tir (&lletra, &numero);     // En genera a lletra i numero una posicio aleatoria valida
        afegir_tir (lletra, numero);    // i la guarda per no tornar-la a repetir
      }
      if (err == 1) {
        fprintf (stderr, "%s: Fem tir %c%i\n", nom_jugador, lletra + offset_lletra, numero + offset_numero);
      }
      // Realitzem tir
      fprintf (stdout, "%c%i\n", lletra + offset_lletra, numero + offset_numero);
      // Esperem resposta
      resposta = toupper( fgetc (stdin) );
      getchar ();               // Eliminem EOL
      switch (resposta) {
        case TOCAT:
          if (!proxim) {
            proxim = TRUE;      //  El vaixell esta proxim a aquesta casella
            direccio = abs ((char) (rand ()) % 4);      // Direccio aleatoria
          }
          afegir_tir_aux (lletra + offset_lletra, numero + offset_numero);
          break;
        case ENFONSAT:
          afegir_tir_aux (lletra + offset_lletra, numero + offset_numero);

          for (i = 0; i < 10; i++)      // Ja em enfonsat el vaixell, podem fer que la taula de tirs
            for (j = 0; j < 10; j++)    // auxiliar pasi a ser la principal.
              tirs[i][j] = tirs_aux[i][j];

          proxim = FALSE;
          encerts++;            // Encerts en serveix per saber si hem enfonsat tots els seus vaixells
          break;                // El torn encara es nostre
        case AIGUA:
          if (proxim) {
            direccio = (direccio + 1) % 4;      // Per aquesta direccio no esta el vaixell
            offset_lletra = offset_numero = 0;
          }
          begin = 0;            // Torn de l'altre
      }
    } else {                    // Torn de l'usuari
      if (err == 1) {
        fprintf (stderr, "%s: Esperem tir\n", nom_jugador);
      }
      // Esperem tir    
      fscanf (stdin, "%s", resposta_tir);
      getchar ();               // Eliminem EOL
      resposta_tir[0] = toupper(resposta_tir[0]);

      // Comprovem si ens ha tocat, enfonsat o ha fet aigua
      c = comprovar_tir (resposta_tir);
      if (err == 1) {
        fprintf (stderr, "%s: Ens han donat a %c\n", nom_jugador, c);
      }

      fprintf (stdout, "%c\n", c);      // Enviem resultat del seu tir
      if (c == AIGUA) {
        begin = 1;              // Torn de l'altre
      }
    }

  } while ((tots_enfonsats () == FALSE) && (encerts < 10));

  if (tots_enfonsats ()) {      // Perdem, ens han enfonsat tots els vaixells
    if (err == 1) {
      fprintf (stderr, "%s: Perd!\n", nom_jugador);
    }
  } else {                      // Guanyem
    if (err == 1) {
      fprintf (stderr, "%s: Guanya!\n", nom_jugador);
    }
  }

  return 0;
}
